﻿//-----------------------------------------------------------------------
// <copyright file="Adapter.cs" company="FRMSc Ltd.">
//     Copyright © FRMSc Ltd. All rights reserved.
// </copyright>
//---

using System;
using System.IO;
using System.Net;
using System.Text;

using System.Runtime.Serialization;
using System.Runtime.Serialization.Json;
using Newtonsoft.Json;
using System.Web;

namespace Safe.API
{
    /// <summary>
    /// Represents the <see cref="Safe.API.TokenAuthenticationAdapter"/> class in the object model.
    /// </summary>
    public class TokenAuthenticationAdapter : IAdapter
    {
        private static readonly string POST = "POST";
        private static readonly string GET = "GET";
        private static readonly string CONTENTTYPE = "application/json";
        private static readonly string FRIENDLYNAMECOOKIE = "FriendlyName";
        private static readonly string UPLOADROSTERURL = "api/UploadRoster";
        private static readonly string UPLOADROSTERBODYFORMAT = "rosterName={0}&rosterData={1}&overwrite={2}&dateFormat={3}&rosterCategories={4}";
        private static readonly string GETRESULTSURL = "api/GetResults";
        private static readonly string GETRESULTSBODYFORMAT = "token={0}&dataKey={1}";        

        private Uri _webApiUrl;
        private Uri _identityUrl;
        private string _clientId;
        private string _clientSecret;

        /// <summary>
        /// Initializes a new instance of the <see cref="Safe.API.TokenAuthenticationAdapter"/> class. 
        /// </summary>
        /// <param name="webApiUrl">The URL to a SAFE site.</param>
        /// <param name="identityUrl">The URL for identity authentication and authorization.</param>
        /// <param name="clientId">The clientId used to authenticate the client.</param>
        /// <param name="clientSecret">The clientSecret used to authenticate the client.</param>
        public TokenAuthenticationAdapter(string webApiUrl, string identityUrl, string clientId, string clientSecret)
        {
            if (string.IsNullOrWhiteSpace(webApiUrl)) throw new NullReferenceException("webApiUrl");
            if (string.IsNullOrWhiteSpace(identityUrl)) throw new NullReferenceException("identityUrl");
            if (string.IsNullOrWhiteSpace(clientId)) throw new NullReferenceException("clientId");
            if (string.IsNullOrWhiteSpace(clientSecret)) throw new NullReferenceException("clientSecret");

            this._webApiUrl = new Uri(webApiUrl);
            this._identityUrl = new Uri(identityUrl);
            this._clientId = clientId;
            this._clientSecret = clientSecret;            

            ServicePointManager.SecurityProtocol = SecurityProtocolType.Tls12;
        }

        #region IAdapter members


        /// <summary>
        /// Uploads a roster to the Safe site, returning a unique token which you can use to check 
        /// the status of processing and gain access to the results.
        /// </summary>
        /// <param name="rosterName">A <see cref="System.String"/> object representing the friendly name for a roster. This will appear in the Safe UI.</param>
        /// <param name="rosterData">A <see cref="System.String"/> object representing a roster CSV blob.</param>  
        /// <param name="key">A <see cref="System.String"/> object representing the API key.</param>  
        /// <param name="overwrite">A <see cref="System.Boolean"/> object representing rosters with same name should be overwritten.</param>  
        /// <param name="rosterCategories">A <see cref="System.String"/> CSV of category names for a roster.</param>        
        /// <returns>A <see cref="System.String"/> object.</returns>
        public string UploadRoster(string rosterName, string rosterData, string key, bool overwrite, string rosterCategories = null)
        {
            var data = new UploadRequest
            {
                RosterName = rosterName,
                RosterData = rosterData,                                
                RosterCategory = rosterCategories,
                Username = _clientId,
                Key = key,
                Overwrite = overwrite
            };

            var request = this.BuildRequest(UPLOADROSTERURL, POST, JsonConvert.SerializeObject(data));

            var token =  this.ReadDataFromResponse(request);         

            return token;
        }

        /// <summary>
        /// Gets the processed data for the roster associated with a given token. Possible return values are:
        /// <para>404 - Token not recognised</para> 
        /// <para>403 - Roster is currently being processed or it completed with an error.</para> 
        /// <para>500 - Unexpected error. Could not find processed roster</para> 
        /// <para>204 - Index value not supported</para> 
        /// <para>A <see cref="System.String"/> object representing the processing status (processing, completedsuccessfully, or completedwitherror)</para> 
        /// <para>A <see cref="System.String"/> object containing any feedback from the processing. This will include any error data.</para> 
        /// <para>A <see cref="System.String"/> object containing a CSV blob representing the processed data.</para> 
        /// </summary>
        /// <param name="token">A <see cref="System.String"/> object representing the unique token.</param>
        /// <param name="dataKey">A <see cref="Safe.API.DataKey"/> object representing the data you want to retrieve.</param>
        /// <returns>A <see cref="System.String"/> object.</returns>
        public string GetResults(string token, DataKey dataKey, string key)
        {            
            var request = this.BuildRequest(
                GETRESULTSURL+ "?token=" + token + "&dataKey=" + dataKey.ToString() + "&key=" + key, GET);
             var response = this.ReadDataFromResponse(request);           
            return response;
        }

        #endregion

        private HttpWebRequest BuildRequest(string actionRelativeUrl, string method, string postBody = null)
        {
            var absoluteUrl = new Uri(this._webApiUrl, actionRelativeUrl);

            var token = Authenticate();

            var request = (HttpWebRequest)WebRequest.Create(absoluteUrl);           
            request.PreAuthenticate = true;
            request.Headers.Add("Authorization", "Bearer " + token);
            request.Method = method;

            if (!string.IsNullOrEmpty(postBody))
            {                
                request.ContentType = CONTENTTYPE;

            using (var stream = request.GetRequestStream())
                {
                    stream.Write(Encoding.UTF8.GetBytes(postBody), 0, postBody.Length);
                }
            }
            else
            {
                request.Method = GET;
            }

            return request;
        }

        private string Authenticate()
        {
            var authUrl = _identityUrl.AbsoluteUri + "connect/token";            

            var plainTextBytes = System.Text.Encoding.UTF8.GetBytes(_clientId + ":" + HttpUtility.UrlEncode(_clientSecret));
            var encodedClientCredentials = System.Convert.ToBase64String(plainTextBytes);

            var authRequest = (HttpWebRequest)WebRequest.Create(authUrl);
            authRequest.Method = POST;
            authRequest.ContentType = "application/x-www-form-urlencoded";
            authRequest.Headers.Add("Authorization", "Basic " + encodedClientCredentials);

            using (var stream = authRequest.GetRequestStream())
            {
                var data = "grant_type=client_credentials&scope=frmsc_safe_upload_api";
                stream.Write(Encoding.UTF8.GetBytes(data), 0, data.Length);
            }

            var authResponse = authRequest.GetResponse();
            AuthResult tokenObj;
            System.IO.Stream responseStream = authResponse.GetResponseStream();
            using (var reader = new System.IO.StreamReader(responseStream))
            {
                var token = reader.ReadToEnd();
                tokenObj = Deserialize<AuthResult>(token);
            }
            responseStream.Close();

            return tokenObj.access_token;
        }

        private string ReadDataFromResponse(HttpWebRequest request)
        {
            var response = request.GetResponse() as HttpWebResponse;

            if (response == null || response.ContentLength == 0)
                throw new NotSupportedException("No response returned.");

            using (var data = new StreamReader(response.GetResponseStream()))
            {
                return data.ReadToEnd();
            }
        }

        private static string Serialize<T>(T obj)
        {
            DataContractJsonSerializer serializer = new DataContractJsonSerializer(obj.GetType());
            MemoryStream ms = new MemoryStream();
            serializer.WriteObject(ms, obj);
            string retVal = Encoding.UTF8.GetString(ms.ToArray());
            return retVal;
        }

        private static T Deserialize<T>(string json)
        {
            T obj = Activator.CreateInstance<T>();
            MemoryStream ms = new MemoryStream(Encoding.Unicode.GetBytes(json));
            DataContractJsonSerializer serializer = new DataContractJsonSerializer(obj.GetType());
            obj = (T)serializer.ReadObject(ms);
            ms.Close();
            return obj;
        }

        [DataContract]
        internal class AuthResult
        {
            [DataMember]
            public string access_token { get; set; }
        }
    }

    public class UploadRequest
    {
        public string RosterName { get; set; }
        public string RosterData { get; set; }
        public string DateFormat { get; set; }
        public string RosterCategory { get; set; }
        public string Username { get; set; }
        public string Key { get; set; }

        public bool Overwrite { get; set; }
    }
}
